# -*- coding: utf-8 -*-
"""
/***************************************************************************
 GraphabProject
                                 A QGIS plugin
 All functions that set and create a Graphab Project
 Generated by Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
                              -------------------
        begin                : 2020-04-20
        git sha              : $Format:%H$
        copyright            : (C) 2020 by ThéMA
        email                : robin.marlinl@gmail.com
        author               : Robin Marlin-Lefebvre
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

import os # This is is needed in the pyqgis console also
import subprocess
import numpy as np
import math

from qgis.PyQt.QtWidgets import QFileDialog, QMessageBox, QInputDialog
from qgis.core import (
    QgsProject,
    QgsVectorLayer,
    QgsRasterLayer,
    QgsSingleBandGrayRenderer,
    QgsRasterBandStats,
    QgsContrastEnhancement,
    QgsVectorLayerJoinInfo,
    QgsFillSymbol,
    QgsField
)

from PyQt5.QtCore import QFileInfo, QVariant, QCoreApplication

# Module to convert graphab xml file to python object
from .xml_to_python.xmlToPython import convertXmlToPy

from .processing.GraphabAlgoProcessing import getGraphabCommand

# Lib use for read correctly path for jointure using uri path
import urllib


class GraphabProject:
    """QGIS Plugin Implementation."""

    def __init__(self, Graphab):
        # Link the object to the main object (Graphab) for shared variables
        self.Graphab = Graphab

    def select_project_file(self):
        # open a window to select a project xlm of Graphab a
        filename, _filter = QFileDialog.getOpenFileName(None, "Select Graphab project","", '*.xml')
        
        # when you close the window but you haven't select a file
        if filename is None or filename == '':
            return

        self.loadProject(filename)

    def loadProject(self, projectFile):
        """It make some verifications, get the XML file
            of a Graphab project, transform it in Python object and load all required files like :
            patches, linkset..."""
        # update the list of loaded projects
        self.updateLoadedProject()

        # if the file is already load the plugin stop
        if os.path.normcase(projectFile) in self.Graphab.tabFileNameProject:
            return
        
        # creation of a python object named 'project' using my lib 'xmlToPython'
        try:
            project = convertXmlToPy(projectFile)
        except:
            # if error stop the plugin
            QMessageBox.warning(None, "Graphab", "Corrupt XML file or representing no Graphab Project")
            return

        projectDir = os.path.dirname(projectFile)
        
        #save new project
        self.Graphab.tabFileNameProject.append(os.path.normcase(projectFile))
        self.Graphab.tabProject.append(project)
        
        # creation of a layers group with the name of the Graphab project
        root = QgsProject.instance().layerTreeRoot()
        projectGroup = root.addGroup(project.name)
        
        csvGroup = projectGroup.addGroup("Tabular data")
        csvGroup.setExpanded(False)
        
        # try to load the source.tif file in 'projectGroup' layers group
        path_to_sourcetif = os.path.join(projectDir, "source.tif")
        self.setSource(path_to_sourcetif, projectGroup)
        
        # try to load the patches.shp file in 'projectGroup' layers group
        path_to_patchesshp = os.path.join(projectDir, "patches.shp")
        self.setPatches(path_to_patchesshp, projectGroup)
                
        # get all graphab linksets and add them in qgis layers group 'Linksets'
        linksetsGroup = projectGroup.insertGroup(0, 'Linksets')
        for linkset in project.costLinks:
            self.setLinkset(linkset.name, linkset, projectDir, linksetsGroup)
            
        # get all graphab graph and add them in qgis layers group 'Graphs' with for each a new layers group
        graphsGroup = projectGroup.insertGroup(0, 'Graphs')
        
        #when you active one graph group; others graph group are disabled
        graphsGroup.setIsMutuallyExclusive(True)
        
        #load each graph
        for graph in project.graphs:
            if graph.type == 1 or graph.type == 2:
                graphGroup = graphsGroup.addGroup(graph.name)
                QgsProject.instance().layerTreeRoot().findGroup(graphGroup.name()).setItemVisibilityChecked(False)
                self.setGraph(projectFile, graphGroup, graph)
            
        # loads patches.csv and link to all patches layer
        self.loadCSV(projectFile, 'patches.csv')

        for linkset in project.costLinks:
            self.reloadLinksetCSV(projectFile, linkset.name)


    #--------------------------------------------------------------------------

    def setSource(self, raster, group):
        """Load the source.tif file of a loaded Project and change its symbology."""
        
        # Get the path and the name of the file to be sure the file can be load
        fileInfo = QFileInfo(raster)
        path = fileInfo.filePath()
        baseName = fileInfo.baseName()

        # Load the file and add it to a specific group to have the same structure of a Graphab project
        layer = QgsRasterLayer(path, baseName)
        QgsProject.instance().addMapLayer(layer, False)
        group.insertLayer(0, layer)

        # If the file is loaded
        if layer.isValid() is True:
            print ("Layer was loaded successfully!")
            # change color of layer that was add
            myGrayRenderer = QgsSingleBandGrayRenderer(layer.dataProvider(), 1)
            layer.setRenderer(myGrayRenderer)
            
            renderer = layer.renderer()
            provider = layer.dataProvider()

            layer_extent = layer.extent()
            uses_band = renderer.usesBands()

            myType = renderer.dataType(uses_band[0])

            stats = provider.bandStatistics(uses_band[0], 
                                            QgsRasterBandStats.All, 
                                            layer_extent, 
                                            0)

            myEnhancement = QgsContrastEnhancement(myType)

            contrast_enhancement = QgsContrastEnhancement.StretchToMinimumMaximum

            myEnhancement.setContrastEnhancementAlgorithm(contrast_enhancement,True)
            myEnhancement.setMinimumValue(stats.minimumValue)
            myEnhancement.setMaximumValue(stats.maximumValue)

            layer.renderer().setContrastEnhancement(myEnhancement)
            # Refresh the symbology
            layer.triggerRepaint()
            
            # Don't set visible the layer such as the loading of a Graphab project
            QgsProject.instance().layerTreeRoot().findLayer(layer.id()).setItemVisibilityChecked(False)

        else:
            print ("Unable to read basename and file path - Your string is probably invalid")

    #--------------------------------------------------------------------------

    def setVoronoi(self, voronoi, group):
        """Laod the Graph voronoi file and change its symbology."""
        
        # Get the path and the name of the file to be sure the file can be load
        path = QFileInfo(voronoi).filePath()

        # Load the file and add it to a specific group to have the same structure of a Graphab project
        layer = QgsVectorLayer(path, "Components")
        QgsProject.instance().addMapLayer(layer, False)
        group.addLayer(layer)

        # If the file is loaded
        if layer.isValid():
            print ("Layer was loaded successfully!")
            
            # Create a new symbology
            symbol = QgsFillSymbol.createSimple({'line_style': 'solid', 'style': 'no'})
         
            # apply symbol to layer renderer
            layer.renderer().setSymbol(symbol)
             
            # repaint the layer
            layer.triggerRepaint()
            self.Graphab.iface.layerTreeView().refreshLayerSymbology(layer.id())
            
            # Don't set visible the layer such as the loading of a Graphab project
            QgsProject.instance().layerTreeRoot().findLayer(layer.id()).setItemVisibilityChecked(False)
            
        else:
            print ("Unable to read basename and file path - Your string is probably invalid")
    
    #--------------------------------------------------------------------------
    
    def setGraph(self, projectPath, group, graph):
        """Load every file that compose a Graph and change symboly of them."""

        # Get the path and the name of the file to be sure the file can be load
        projectDir = os.path.dirname(projectPath);
        path_to_patchesshp = os.path.join(projectDir, "patches.shp")
        pathPatches = QFileInfo(path_to_patchesshp).filePath()
        
        # Load the file and add it to a specific group to have the same structure of a Graphab project
        layer = QgsVectorLayer(pathPatches, "Nodes")
        QgsProject.instance().addMapLayer(layer, False)
        group.addLayer(layer)

        # If the file is loaded
        if layer.isValid():
            print ("Layer was loaded successfully!")
            
            topoLinkset = os.path.join(projectDir, graph.name + '-topo_links.shp')
            if not os.path.isfile(topoLinkset):
                self.createTopoLinkset(projectPath, graph)
            
            # Load Linkset with filter or not depending of the type of the Graph loaded
            if graph.type == 1:
                self.setLinkset("Edges", graph.costLink, projectDir, group, graph.name, True)
                self.setLinkset("Paths", graph.costLink, projectDir, group, None,  False)
                
            if graph.type == 2:
                self.setLinkset("Edges", graph.costLink, projectDir, group, graph.name, True)
                self.setLinkset("Paths", graph.costLink, projectDir, group, None, False, graph.threshold)
            
            # Load voronoi file
            path_to_voronoi = os.path.join(projectDir, graph.name+"-voronoi.shp")
            self.setVoronoi(path_to_voronoi, group)
            
            # Set default Graph symbology parameters
            fieldname = 'Capacity'
            styleName = 'BleuDegrade'
            minScale = None
            maxScale = None
            defaultSize = True

            # Change the symbology of the current Graph
            self.Graphab.GraphabStyle.prepareStyle(fieldname, layer, styleName, minScale, maxScale, defaultSize)

            # Don't set visible the layer such as the loading of a Graphab project
            QgsProject.instance().layerTreeRoot().findLayer(layer.id()).setItemVisibilityChecked(True)
            
            # Collapse the layer group
            group.setExpanded(False)

            return layer
        else:
            print ("Unable to read basename and file path - Your string is probably invalid")
            return None
            

    #--------------------------------------------------------------------------
    
    def setPatches(self, vector, group):
        """Load patches.shp file and change its symbology."""

        # Get the path and the name of the file to be sure the file can be load
        fileInfo = QFileInfo(vector)
        path = fileInfo.filePath()
        baseName = fileInfo.baseName()

        # Load the file and add it to a specific group to have the same structure of a Graphab project
        layer = QgsVectorLayer(path, baseName)
        QgsProject.instance().addMapLayer(layer, False)
        group.insertLayer(0, layer)

        # If the file is loaded
        if layer.isValid() is True:
            print ("Layer was loaded successfully!")
            
            # Change the symbology in a green
            self.Graphab.GraphabStyle.colorizeFill(layer, '#65A252')
            
            return layer
        else:
            print ("Unable to read basename and file path - Your string is probably invalid")
            return None

    #--------------------------------------------------------------------------
    
    def setLinkset(self, name, linkset, projectDir, group, graphName = None, visible = False, threshold = 0):
        """Load Linkset and change symbology depending on their type."""

        if graphName:
            linksetFile = graphName+"-topo_links.shp"
        else:
            linksetFile = linkset.name+"-links.shp"

        # Get the path and the name of the file to be sure the file can be load
        fileInfo = QFileInfo(os.path.join(projectDir, linksetFile))
        path = fileInfo.filePath()

        # Load the file and add it to a specific group to have the same structure of a Graphab project
        layer = QgsVectorLayer(path, name, "ogr")
        QgsProject.instance().addMapLayer(layer, False)
        group.addLayer(layer)

        # If the file is loaded
        if layer.isValid() is True:
            print ("Layer was loaded successfully!")
                
            # Set up of the color which depends of the type of the linkset
            if(linkset.type) == 1:
                self.Graphab.GraphabStyle.colorizeLine(layer, '#B8C45D')
            elif(linkset.type) == 2:
                self.Graphab.GraphabStyle.colorizeLine(layer, '#25372B')
            else:
                self.Graphab.GraphabStyle.colorizeLine(layer, '#BCC3AC')
                
            if threshold > 0:
                #set filter ("Dist"<Graph.threshold)
                equation = '"Dist" <= %d' % (threshold,)
                layer.setSubsetString(equation)
            
            # Don't set visible the layer such as the loading of a Graphab project
            QgsProject.instance().layerTreeRoot().findLayer(layer.id()).setItemVisibilityChecked(visible)
            
        else:
            print ("Unable to read basename and file path - Your string is probably invalid")
            
    #--------------------------------------------------------------------------
    
    def addConcatenateField(self, layer, fieldName, firstField, secondField, separator):
        """Create a virtual field on a layer to allow to join csv with concatenate ID"""
        # add apostrophes at the beginning and the end of the separator to be sure it was count as string
        separator = "'" + separator + "'"
        
        # create virtual field to concatenate 2 fields
        field = QgsField( fieldName, QVariant.String )
        layer.addExpressionField( 'concat("' + firstField + '", ' + separator + ', "' + secondField + '")', field )
    

    #--------------------------------------------------------------------------
    
    def updateLoadedProject(self):
        """Compare the name of Project groups load in the QGIS layer window and the name of Projects saved. 
        Delete Projects saved they are missing."""
        
        # To stock index of project missing
        tabToDelete = []
        index = 0
        # Get QGIS layer window
        root = QgsProject.instance().layerTreeRoot()
        
        # For each Project saved
        for project in self.Graphab.tabProject:
            index += 1
            # Try to find the Project group
            projectGroup = root.findGroup(project.name)
            # If Project group is missing
            if projectGroup == None:
                tabToDelete.append(index)
        
        # Begin by the end of index to provide index error and delete Project + File project
        tabToDelete.reverse()
        for toDelete in tabToDelete:
            self.Graphab.tabProject.pop(toDelete - 1)
            self.Graphab.tabFileNameProject.pop(toDelete - 1)
            self.Graphab.tabListPatches.pop(toDelete - 1)

    # --------------------------------------------------------------------------

    def loadCSV(self, projectPath, csvName):

        # find if the project is already loaded
        indexProject = self.indexOfProject(projectPath)

        # if the project is not loaded nothing to do
        if indexProject < 0:
            return
        isPatch = csvName == 'patches.csv'

        # get the name of the project and its group
        project = self.Graphab.tabProject[indexProject]
        prjGroup = QgsProject.instance().layerTreeRoot().findGroup(project.name)
        csvGroup = prjGroup.findGroup("Tabular data")

        # find and delete the csv file and remove automatically all joins
        for layerGraphTree in csvGroup.children():
            if layerGraphTree.name() == csvName:
                QgsProject.instance().removeMapLayer(layerGraphTree.layer())
                break

        # get the path to the directory of the project and the csv path for patches
        path_to_csv = os.path.join(os.path.dirname(projectPath), csvName)

        if not os.path.isfile(path_to_csv):
            return

        # To make the uri path correct for special character we use urllib.parse.quote() function
        csv_file_path = urllib.parse.quote(path_to_csv)

        # Find and load the CSV file
        csv_uri = 'file:///' + csv_file_path + '?watchFile=no'
        csvLayer = QgsVectorLayer(csv_uri, csvName, "delimitedtext")

        if not isPatch:
            self.addConcatenateField(csvLayer, 'Id', 'ID1', 'ID2', '-')

        # Add it to the QGIS layer window
        QgsProject.instance().addMapLayer(csvLayer, False)
        csvGroup.addLayer(csvLayer)

        # Create the link between the CSV file and the layer based on a specific field ('Id' in general)
        joinObject = QgsVectorLayerJoinInfo()
        joinObject.setJoinLayerId(csvLayer.id())
        joinObject.setJoinFieldName('Id')
        joinObject.setTargetFieldName('Id')
        joinObject.setUsingMemoryCache(True)
        if isPatch:
            joinObject.setJoinFieldNamesBlackList(['Area', 'Perim'])
        else:
            joinObject.setJoinFieldNamesBlackList(['ID1', 'ID2', 'Dist', 'DistM'])
        joinObject.setPrefix(self.Graphab.prefix)
        joinObject.setJoinLayer(csvLayer)

        shpName = csvName.replace('.csv', '.shp')
        shpNameTopo = []
        if shpName.endswith('-links.shp'):
            for graph in project.graphs:
                if graph.costLink.name == csvName.replace('-links.csv', ''):
                    shpNameTopo.append(graph.name + '-topo_links.shp')

        for l in prjGroup.findLayers():
            layer = l.layer()
            if layer.source().endswith(shpName) or (shpName + '|') in layer.source():
                layer.addJoin(joinObject)
            # for graph topo linkset
            for name in shpNameTopo:
                if layer.source().endswith(name):
                    layer.addJoin(joinObject)


    # --------------------------------------------------------------------------
    def reloadLinksetCSV(self, projectPath, linksetName):
        self.loadCSV(projectPath, linksetName + '-links.csv')

    #--------------------------------------------------------------------------

    def reloadPatchCSV(self, projectPath):
        """Reload project csv layers using for patches by deleting and reloading them and return
        the project name to confirm or if it doesn't exist just return None"""

        self.loadCSV(projectPath, 'patches.csv')

    #--------------------------------------------------------------------------
    
    def indexOfProject(self, projectPath):
        """Function that return the index of the project represented by the given path if the
        project is already loaded, else return -1"""
        
        #update loaded project before
        self.updateLoadedProject()
        try:
            return self.Graphab.tabFileNameProject.index(projectPath)
        except ValueError:
            return -1

    # --------------------------------------------------------------------------

    def getProjectGroup(self, projectPath):
        """Function that return the index of the project represented by the given path if the
        project is already loaded, else return -1"""
        # find if the project is already loaded
        indexProject = self.indexOfProject(projectPath)

        # if the project is loaded, add the new Graph
        if indexProject >= 0:
            # replace the previous stocked project by the new one
            result = self.upgradeLoadedProject(indexProject)
            if result == -1:
                return None

            # get theproject name
            nameProject = self.Graphab.tabProject[indexProject].name

            # Try to find the Project group
            return QgsProject.instance().layerTreeRoot().findGroup(nameProject)
        else:
            return None

    #--------------------------------------------------------------------------
    
    def deleteJoins(self, indexProject, csvGroup):
        """Delete all joins on patches layer saved for one project and the csv file"""
        # Delete all joins of all patches layer saved
        for layer in self.Graphab.tabListPatches[indexProject]:
            # get all joins of one layer
            joinsInfo = layer.vectorJoins()
            for joinInfo in joinsInfo:
                # delete the join and make a verification that works
                ok = layer.removeJoin(joinInfo.joinLayerId())
                if ok:
                    print("Join removed successfully")
                else:
                    print("Couldn't remove the layer join")

        # find and delete the csv file
        for layerGraphTree in csvGroup.children():
            if layerGraphTree.name() == 'patches.csv':
                QgsProject.instance().removeMapLayer(layerGraphTree.layer())
                break
    
    #--------------------------------------------------------------------------
    
    def addGraphProject(self, graphName, projectPath):
        """This function load a Graph that had been created when the project is already loaded"""
        
        # find if the project is already loaded
        indexProject = self.indexOfProject(projectPath)
            
        # if the project is loaded, add the new Graph
        if indexProject >= 0:
            # replace the previous stocked project by the new one
            result = self.upgradeLoadedProject(indexProject)
            if result == -1:
                return
            
            # get theproject name
            nameProject = self.Graphab.tabProject[indexProject].name

            # Try to find the Project group
            projectGroup = QgsProject.instance().layerTreeRoot().findGroup(nameProject)
            
            # get 'Graphs' group
            graphsGroup = projectGroup.children()[0]

            # get the project with the new graph
            project = self.Graphab.tabProject[indexProject]

            # try to find the new graph with its name and load it
            for graph in project.graphs:
                if graph.name == graphName:
                    graphGroup = graphsGroup.addGroup(graph.name)
                    self.setGraph(projectPath, graphGroup, graph)
                    self.reloadPatchCSV(projectPath)
                    self.reloadLinksetCSV(projectPath, graph.costLink.name)

    
    #--------------------------------------------------------------------------
    
    def addLinksetProject(self, linksetName, projectPath):
        """This function load a Linkset that had been created when the project is already loaded"""
        
        # find if the project is already loaded
        indexProject = self.indexOfProject(projectPath)
        if indexProject == -1:
            return

        # replace the previous stocked project by the new one
        result = self.upgradeLoadedProject(indexProject)
        if result == -1:
            return

        # get theproject name
        nameProject = self.Graphab.tabProject[indexProject].name

        # Try to find the Project group
        projectGroup = QgsProject.instance().layerTreeRoot().findGroup(nameProject)

        # get 'Linksets' group
        linksetsGroup = projectGroup.children()[1]

        # get the project with the new linkset
        project = self.Graphab.tabProject[indexProject]

        # try to find the new linkset with its name and load it
        for linkset in project.costLinks:
            if linkset.name == linksetName:
                self.setLinkset(linksetName, linkset, os.path.dirname(projectPath), linksetsGroup, visible=True)

        self.reloadLinksetCSV(projectPath, linksetName)
    
    #--------------------------------------------------------------------------
    
    def upgradeLoadedProject(self, indexProject):
        """This function replace old project by new project after a project modif"""
        
        # create the new project
        try:
            projectTemp = convertXmlToPy(self.Graphab.tabFileNameProject[indexProject])
        except:
            # if error stop the plugin
            print("Corrupt XML file or representing no Graphab Project")
            return -1
        
        self.project = projectTemp
        
        # replace old project by new project
        self.Graphab.tabProject[indexProject] = self.project
        
        return 0


    def createTopoLinkset(self, projectFile, graph):
        cmd = getGraphabCommand() + ['--project', projectFile, '--usegraph', graph.name, '--topo']
        return subprocess.call(cmd) == 0

    def removeLinkset(self, projectFile, linksetName):
        cmd = getGraphabCommand() + ['--project', projectFile, '--removelinkset' + linksetName]
        ret = subprocess.call(cmd)
        if ret == 0:
            self.upgradeLoadedProject(self.indexOfProject(projectFile))
            return True
        else:
            return False

    def removeGraph(self, projectFile, graphName):
        cmd = getGraphabCommand() + ['--project', projectFile, '--removegraph' + graphName]
        ret = subprocess.call(cmd)
        if ret == 0:
            self.upgradeLoadedProject(self.indexOfProject(projectFile))
            return True
        else:
            return False

    # --------------------------------------------------------------------------

    def convertDistance(self, indexProject, linksetName):

        d, ok = QInputDialog().getDouble(self.Graphab.iface.mainWindow(), "Distance conversion",
                                         QCoreApplication.translate('py', "Distance (in meter) :"), 1000)
        if ok and d:
            # get theproject name
            nameProject = self.Graphab.tabProject[indexProject].name
            # get 'Linksets' group
            linksetsGroup = QgsProject.instance().layerTreeRoot().findGroup(nameProject).children()[1]
            for layer in linksetsGroup.findLayers():
                if layer.name() == linksetName:
                    linksetLayer = layer.layer()

            dist = []
            cost = []
            for feature in linksetLayer.getFeatures():
                dist.append(math.log(feature.attribute("DistM")))
                cost.append(math.log(feature.attribute("Dist")))

            a, b = np.polyfit(dist, cost, deg=1)

            return math.exp(a*math.log(d)+b)
        else:
            return None