# -*- coding: utf-8 -*-

"""
/***************************************************************************
 GraphabAlgoProcessing
                                 A QGIS plugin
 Execute command send by its Children
 Generated by Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
                              -------------------
        begin                : 2020-07-02
        copyright            : (C) 2020 by ThéMA
        email                : robin.marlinl@gmail.com
        author               : Robin Marlin-Lefebvre
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

__author__ = 'robin.marlinl@gmail.com'
__date__ = '2020-07-02'
__copyright__ = '(C) 2020 by ThéMA'

# This will get replaced with a git SHA1 when you do a git archive

__revision__ = '$Format:%H$'

import io
import sys
from builtins import Exception, IOError

from PyQt5.QtCore import QCoreApplication, QFile, QUrl, QIODevice, QEventLoop
from PyQt5.QtNetwork import QNetworkRequest, QNetworkAccessManager
from qgis.core import (
    QgsProcessingAlgorithm,
    QgsMessageLog,
    Qgis, QgsNetworkAccessManager, QgsFileDownloader
)

import subprocess
import os

class GraphabAlgoProcessing(QgsProcessingAlgorithm):
    """
    This Processing algorithms should extend the QgsProcessingAlgorithm
    class because it's the main file that execute command.
    """
    def __init__(self, Graphab = None):
        super().__init__()
        
        self.executionError = False
        self.Graphab = Graphab
        self.loglines = []

    def group(self):
        """
        Returns the name of the group this algorithm belongs to. This string
        should be localised.
        """
        return 'Graphab'

    def groupId(self):
        """
        Returns the unique ID of the group this algorithm belongs to. This
        string should be fixed for the algorithm, and must not be localised.
        The group id should be unique within each provider. Group id should
        contain lowercase alphanumeric characters only and no spaces or other
        formatting characters.
        """
        return 'graphab'

    def translate(self, context, string):
        return QCoreApplication.translate(context, string)
        
    def processAlgorithm(self, parameters, context, feedback):
        """
        Execute a given command by adding the java executable at the beginning. Get the output
        of the execution and stock it to show it after the execution. Depending of the command
        this function get important values to make post execution like loading a new Linkset or Graph.
        """

        fused_command = getGraphabCommand() + parameters
        QgsMessageLog.logMessage(str(fused_command), 'Processing', Qgis.Info)
        feedback.pushInfo('Graphab command:\n')
        feedback.pushCommandInfo(str(fused_command) + "\n\n")
        feedback.pushInfo('Graphab command output:\n')

        self.loglines = []
        self.loglines.append('Graphab execution console output : ')

        with subprocess.Popen(
                    fused_command,
                    stdout=subprocess.PIPE,
                    stdin=subprocess.DEVNULL,
                    stderr=subprocess.STDOUT
                ) as proc:
            reader = io.TextIOWrapper(proc.stdout, errors='ignore', encoding='utf8')
            p = 0
            line = reader.readline()
            while line != '':
                self.loglines.append(line)
                if line.endswith(": 0,0%\n") or line.endswith(": 0.0%\n"):
                    feedback.pushConsoleInfo(line[0:-7])
                    p = 0
                elif line.endswith("%\n"):
                    p += 0.1
                    feedback.setProgress(p)
                else:
                    feedback.pushConsoleInfo(line[0:-1])
                if feedback.isCanceled():
                    feedback.pushConsoleInfo("Try to cancel Graphab...")
                    proc.kill()

                line = reader.readline()

            if proc.wait() > 0:
                # raise an exception
                self.executionError = True
                raise IOError("Error during Graphab execution!\n" + ''.join(self.loglines))
            else:
                if feedback.isCanceled():
                    self.executionError = True
                    feedback.pushInfo("GRAPHAB EXECUTION CANCELED\n\n\n")
                else:
                    self.executionError = False
                    feedback.pushInfo("GRAPHAB EXECUTION FINISHED WITH SUCCESS!\n\n\n")


        QgsMessageLog.logMessage(''.join(self.loglines), 'Processing', Qgis.Info)
        
        return {}


def getGraphabCommand():
    inputJar = os.path.normcase(r''+os.path.join(os.path.dirname(os.path.realpath(__file__)), 'graphab-2.6.jar'))
    jarFile = QFile(inputJar)
    if not jarFile.exists():
        downloadGraphab(inputJar)
        if not jarFile.exists():
            raise IOError("Impossible to download Graphab software. Check log messages")

    return [getJavaCommand(), "-Djava.awt.headless=true", "-jar", inputJar]

def downloadGraphab(jarFile):
    url = QUrl("http://thema.univ-fcomte.fr/productions/download.php?name=graphab&version=2.6&username=Graph4QGis&institution=Python")

    def downloadFailed(error):
        errorMessage = "Impossible to download Graphab software, check your internet connexion.\nDetails : ".join(error)
        QgsMessageLog.logMessage(errorMessage, 'Processing', Qgis.Critical)

    loop = QEventLoop()
    downloader = QgsFileDownloader(url, jarFile, delayStart=True)

    # connecting signals to functions
    downloader.downloadExited.connect(loop.quit)
    downloader.downloadError.connect(downloadFailed)

    downloader.startDownload()
    loop.exec_()


def getJavaCommand():
    java = 'java'
    if sys.platform == 'win32':
        java = 'javaw.exe'

    java_path = os.getenv("JAVA_HOME")
    if java_path != None:
        return os.path.join(java_path, 'bin', java)
    else:
        if sys.platform == 'win32':
            cmd = getJavaWin(java)
            if cmd:
                return cmd
            else:
                for java_dir in ["Java", "AdoptOpenJDK"]:
                    java_path = "C:\\Program Files\\" + java_dir
                    if os.path.isdir(java_path):
                        for jdir in os.listdir(java_path):
                            javaExec = os.path.join(java_path, jdir, "bin", java)
                            if os.path.isfile(javaExec):
                                return javaExec
        return java

def getJavaWin(program):
    import os
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    for path in getJavaHomesWin():
        path = path.strip('"')
        exe_file = os.path.join(path, 'bin', program)
        if is_exe(exe_file):
            return exe_file

    for path in getenv_system("PATH").split(os.pathsep):
        path = path.strip('"')
        exe_file = os.path.join(path, program)
        if is_exe(exe_file):
            return exe_file
    return None

# This function gets a system variable
# it was necessary to use this instead of os.environ["PATH"] because QGIS overwrites the path variable
# the win32 libraries appear not to be part of the standard python install, but they are included in the
# python version that comes with QGIS
def getenv_system(varname):
    v = ''
    try:
        import win32api
        import win32con
    except:
        return v

    try:
        rkey = win32api.RegOpenKey(win32con.HKEY_LOCAL_MACHINE, 'SYSTEM\\CurrentControlSet\\Control\\Session Manager\\Environment')
        try:
            v = str(win32api.RegQueryValueEx(rkey, varname)[0])
            v = win32api.ExpandEnvironmentStrings(v)
        except:
            pass
    finally:
        win32api.RegCloseKey(rkey)
    return v

def getJavaHomesWin():

    javahomes = []
    try:
        import win32api
        import win32con
    except:
        return javahomes

    try:
        rkey = win32api.RegOpenKey(win32con.HKEY_LOCAL_MACHINE, 'SOFTWARE\\JavaSoft\\Java Runtime Environment')
        versions = win32api.RegEnumKeyEx(rkey)
        for version in versions:
            try:
                vkey = win32api.RegOpenKey(win32con.HKEY_LOCAL_MACHINE,
                                           'SOFTWARE\\JavaSoft\\Java Runtime Environment\\' + version[0])
                javahomes.append(str(win32api.RegQueryValueEx(vkey, 'JavaHome')[0]))
                win32api.RegCloseKey(vkey)
            except:
                pass
        win32api.RegCloseKey(rkey)
    except:
        pass

    return javahomes